@@ -4,10 +4,16 @@ import (
 	"bytes"
 	"fmt"
 	"strconv"
+
+	"github.com/tdewolff/parse/v2"
 )
 
 var ErrInvalidJSON = fmt.Errorf("invalid JSON")
 
+type JSONer interface {
+	JSON(*bytes.Buffer) error
+}
+
 // AST is the full ECMAScript abstract syntax tree.
 type AST struct {
 	Comments  [][]byte // first comments in file
@@ -86,12 +92,8 @@ func (v Var) JS() string {
 	return v.String()
 }
 
-// JSON converts the node back to valid JSON
-func (n Var) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // VarsByUses is sortable by uses in descending order.
+// TODO: write custom sorter for varsbyuses
 type VarsByUses VarArray
 
 func (vs VarsByUses) Len() int {
@@ -129,11 +131,11 @@ func (vs VarArray) String() string {
 
 // Scope is a function or block scope with a list of variables declared and used.
 type Scope struct {
-	Parent, Func   *Scope   // Parent is nil for global scope, Parent equals Func for function scope
+	Parent, Func   *Scope   // Parent is nil for global scope
 	Declared       VarArray // Link in Var are always nil
 	Undeclared     VarArray
-	NumVarDecls    uint16 // number of variable declaration statements in a function scope
-	NumForInit     uint16 // offset into Declared to mark variables used in for initializer
+	VarDecls       []*VarDecl
+	NumForDeclared uint16 // offset into Declared to mark variables used in for statements
 	NumArguments   uint16 // offset into Undeclared to mark variables used in arguments
 	IsGlobalOrFunc bool
 	HasWith        bool
@@ -170,7 +172,7 @@ func (s *Scope) Declare(decl DeclType, name []byte) (*Var, bool) {
 		}
 		v.Uses++
 		for s != curScope {
-			curScope.addUndeclared(v) // add variable declaration as used variable to the current scope
+			curScope.AddUndeclared(v) // add variable declaration as used variable to the current scope
 			curScope = curScope.Parent
 		}
 		return v, true
@@ -198,7 +200,7 @@ func (s *Scope) Declare(decl DeclType, name []byte) (*Var, bool) {
 	v.Uses++
 	s.Declared = append(s.Declared, v)
 	for s != curScope {
-		curScope.addUndeclared(v) // add variable declaration as used variable to the current scope
+		curScope.AddUndeclared(v) // add variable declaration as used variable to the current scope
 		curScope = curScope.Parent
 	}
 	return v, true
@@ -222,11 +224,11 @@ func (s *Scope) Use(name []byte) *Var {
 }
 
 // findDeclared finds a declared variable in the current scope.
-func (s *Scope) findDeclared(name []byte, skipForInit bool) *Var {
+func (s *Scope) findDeclared(name []byte, skipForDeclared bool) *Var {
 	start := 0
-	if skipForInit {
+	if skipForDeclared {
 		// we skip the for initializer for declarations (only has effect for let/const)
-		start = int(s.NumForInit)
+		start = int(s.NumForDeclared)
 	}
 	// reverse order to find the inner let first in `for(let a in []){let a; {a}}`
 	for i := len(s.Declared) - 1; start <= i; i-- {
@@ -251,7 +253,7 @@ func (s *Scope) findUndeclared(name []byte) *Var {
 }
 
 // add undeclared variable to scope, this is called for the block scope when declaring a var in it
-func (s *Scope) addUndeclared(v *Var) {
+func (s *Scope) AddUndeclared(v *Var) {
 	// don't add undeclared symbol if it's already there
 	for _, vorig := range s.Undeclared {
 		if v == vorig {
@@ -261,9 +263,10 @@ func (s *Scope) addUndeclared(v *Var) {
 	s.Undeclared = append(s.Undeclared, v) // add variable declaration as used variable to the current scope
 }
 
-// MarkForInit marks the declared variables in current scope as for statement initializer to distinguish from declarations in body.
-func (s *Scope) MarkForInit() {
-	s.NumForInit = uint16(len(s.Declared))
+// MarkForStmt marks the declared variables in current scope as for statement initializer to distinguish from declarations in body.
+func (s *Scope) MarkForStmt() {
+	s.NumForDeclared = uint16(len(s.Declared))
+	s.MarkArguments() // ensures for different b's in for(var a in b){let b}
 }
 
 // MarkArguments marks the undeclared variables in the current scope as function arguments. It ensures different b's in `function f(a=b){var b}`.
@@ -336,7 +339,6 @@ func (s *Scope) Unscope() {
 type INode interface {
 	String() string
 	JS() string
-	JSON() (string, error)
 }
 
 // IStmt is a dummy interface for statements.
@@ -388,14 +390,6 @@ func (n BlockStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n BlockStmt) JSON() (string, error) {
-	if len(n.List) != 1 {
-		return "", ErrInvalidJSON
-	}
-	return n.List[0].JSON()
-}
-
 // EmptyStmt is an empty statement.
 type EmptyStmt struct {
 }
@@ -409,11 +403,6 @@ func (n EmptyStmt) JS() string {
 	return ";"
 }
 
-// JSON converts the node back to valid JSON
-func (n EmptyStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ExprStmt is an expression statement.
 type ExprStmt struct {
 	Value IExpr
@@ -432,11 +421,6 @@ func (n ExprStmt) JS() string {
 	return n.Value.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n ExprStmt) JSON() (string, error) {
-	return n.Value.JSON()
-}
-
 // IfStmt is an if statement.
 type IfStmt struct {
 	Cond IExpr
@@ -472,11 +456,6 @@ func (n IfStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n IfStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // DoWhileStmt is a do-while iteration statement.
 type DoWhileStmt struct {
 	Cond IExpr
@@ -499,11 +478,6 @@ func (n DoWhileStmt) JS() string {
 	return s + " while (" + n.Cond.JS() + ")"
 }
 
-// JSON converts the node back to valid JSON
-func (n DoWhileStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // WhileStmt is a while iteration statement.
 type WhileStmt struct {
 	Cond IExpr
@@ -523,11 +497,6 @@ func (n WhileStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n WhileStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ForStmt is a regular for iteration statement.
 type ForStmt struct {
 	Init IExpr // can be nil
@@ -538,7 +507,7 @@ type ForStmt struct {
 
 func (n ForStmt) String() string {
 	s := "Stmt(for"
-	if n.Init != nil {
+	if v, ok := n.Init.(*VarDecl); !ok && n.Init != nil || ok && len(v.List) != 0 {
 		s += " " + n.Init.String()
 	}
 	s += " ;"
@@ -555,7 +524,7 @@ func (n ForStmt) String() string {
 // JS converts the node back to valid JavaScript
 func (n ForStmt) JS() string {
 	s := "for ("
-	if n.Init != nil {
+	if v, ok := n.Init.(*VarDecl); !ok && n.Init != nil || ok && len(v.List) != 0 {
 		s += n.Init.JS()
 	} else {
 		s += " "
@@ -571,11 +540,6 @@ func (n ForStmt) JS() string {
 	return s + ") " + n.Body.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n ForStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ForInStmt is a for-in iteration statement.
 type ForInStmt struct {
 	Init  IExpr
@@ -592,11 +556,6 @@ func (n ForInStmt) JS() string {
 	return "for (" + n.Init.JS() + " in " + n.Value.JS() + ") " + n.Body.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n ForInStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ForOfStmt is a for-of iteration statement.
 type ForOfStmt struct {
 	Await bool
@@ -622,11 +581,6 @@ func (n ForOfStmt) JS() string {
 	return s + " (" + n.Init.JS() + " of " + n.Value.JS() + ") " + n.Body.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n ForOfStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // CaseClause is a case clause or default clause for a switch statement.
 type CaseClause struct {
 	TokenType
@@ -660,11 +614,6 @@ func (n CaseClause) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n CaseClause) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // SwitchStmt is a switch statement.
 type SwitchStmt struct {
 	Init IExpr
@@ -689,11 +638,6 @@ func (n SwitchStmt) JS() string {
 	return s + " }"
 }
 
-// JSON converts the node back to valid JSON
-func (n SwitchStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // BranchStmt is a continue or break statement.
 type BranchStmt struct {
 	Type  TokenType
@@ -717,11 +661,6 @@ func (n BranchStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n BranchStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ReturnStmt is a return statement.
 type ReturnStmt struct {
 	Value IExpr // can be nil
@@ -744,11 +683,6 @@ func (n ReturnStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n ReturnStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // WithStmt is a with statement.
 type WithStmt struct {
 	Cond IExpr
@@ -764,11 +698,6 @@ func (n WithStmt) JS() string {
 	return "with (" + n.Cond.JS() + ") " + n.Body.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n WithStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // LabelledStmt is a labelled statement.
 type LabelledStmt struct {
 	Label []byte
@@ -784,11 +713,6 @@ func (n LabelledStmt) JS() string {
 	return string(n.Label) + ": " + n.Value.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n LabelledStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ThrowStmt is a throw statement.
 type ThrowStmt struct {
 	Value IExpr
@@ -803,11 +727,6 @@ func (n ThrowStmt) JS() string {
 	return "throw " + n.Value.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n ThrowStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // TryStmt is a try statement.
 type TryStmt struct {
 	Body    *BlockStmt
@@ -847,11 +766,6 @@ func (n TryStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n TryStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // DebuggerStmt is a debugger statement.
 type DebuggerStmt struct {
 }
@@ -865,11 +779,6 @@ func (n DebuggerStmt) JS() string {
 	return "debugger"
 }
 
-// JSON converts the node back to valid JSON
-func (n DebuggerStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // Alias is a name space import or import/export specifier for import/export statements.
 type Alias struct {
 	Name    []byte // can be nil
@@ -889,11 +798,6 @@ func (alias Alias) JS() string {
 	return alias.String()
 }
 
-// JSON converts the node back to valid JSON
-func (n Alias) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ImportStmt is an import statement.
 type ImportStmt struct {
 	List    []Alias
@@ -958,11 +862,6 @@ func (n ImportStmt) JS() string {
 	return s + " " + string(n.Module)
 }
 
-// JSON converts the node back to valid JSON
-func (n ImportStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ExportStmt is an export statement.
 type ExportStmt struct {
 	List    []Alias
@@ -1026,11 +925,6 @@ func (n ExportStmt) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n ExportStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // DirectivePrologueStmt is a string literal at the beginning of a function or module (usually "use strict").
 type DirectivePrologueStmt struct {
 	Value []byte
@@ -1045,11 +939,6 @@ func (n DirectivePrologueStmt) JS() string {
 	return string(n.Value)
 }
 
-// JSON converts the node back to valid JSON
-func (n DirectivePrologueStmt) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 func (n BlockStmt) stmtNode()             {}
 func (n EmptyStmt) stmtNode()             {}
 func (n ExprStmt) stmtNode()              {}
@@ -1110,11 +999,6 @@ func (n PropertyName) JS() string {
 	return n.String()
 }
 
-// JSON converts the node back to valid JSON
-func (n PropertyName) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // BindingArray is an array binding pattern.
 type BindingArray struct {
 	List []BindingElement
@@ -1156,11 +1040,6 @@ func (n BindingArray) JS() string {
 	return s + "]"
 }
 
-// JSON converts the node back to valid JSON
-func (n BindingArray) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // BindingObjectItem is a binding property.
 type BindingObjectItem struct {
 	Key   *PropertyName // can be nil
@@ -1188,11 +1067,6 @@ func (n BindingObjectItem) JS() string {
 	return " " + n.Value.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n BindingObjectItem) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // BindingObject is an object binding pattern.
 type BindingObject struct {
 	List []BindingObjectItem
@@ -1244,11 +1118,6 @@ func (n BindingObject) JS() string {
 	return s + " }"
 }
 
-// JSON converts the node back to valid JSON
-func (n BindingObject) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // BindingElement is a binding element.
 type BindingElement struct {
 	Binding IBinding // can be nil (in case of ellision)
@@ -1278,11 +1147,6 @@ func (n BindingElement) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n BindingElement) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 func (v *Var) bindingNode()          {}
 func (n BindingArray) bindingNode()  {}
 func (n BindingObject) bindingNode() {}
@@ -1292,7 +1156,9 @@ func (n BindingObject) bindingNode() {}
 // VarDecl is a variable statement or lexical declaration.
 type VarDecl struct {
 	TokenType
-	List []BindingElement
+	List             []BindingElement
+	Scope            *Scope
+	InFor, InForInOf bool
 }
 
 func (n VarDecl) String() string {
@@ -1315,11 +1181,6 @@ func (n VarDecl) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n VarDecl) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // Params is a list of parameters for functions, methods, and arrow function.
 type Params struct {
 	List []BindingElement
@@ -1361,11 +1222,6 @@ func (n Params) JS() string {
 	return s + ")"
 }
 
-// JSON converts the node back to valid JSON
-func (n Params) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // FuncDecl is an (async) (generator) function declaration or expression.
 type FuncDecl struct {
 	Async     bool
@@ -1408,11 +1264,6 @@ func (n FuncDecl) JS() string {
 	return s + " " + n.Params.JS() + " " + n.Body.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n FuncDecl) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // MethodDecl is a method definition in a class declaration.
 type MethodDecl struct {
 	Static    bool
@@ -1468,11 +1319,6 @@ func (n MethodDecl) JS() string {
 	return s[1:]
 }
 
-// JSON converts the node back to valid JSON
-func (n MethodDecl) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // FieldDefinition is a field definition in a class declaration.
 type FieldDefinition struct {
 	Name PropertyName
@@ -1496,11 +1342,6 @@ func (n FieldDefinition) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n FieldDefinition) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ClassDecl is a class declaration.
 type ClassDecl struct {
 	Name        *Var  // can be nil
@@ -1545,11 +1386,6 @@ func (n ClassDecl) JS() string {
 	return s + "}"
 }
 
-// JSON converts the node back to valid JSON
-func (n ClassDecl) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 func (n VarDecl) stmtNode()   {}
 func (n FuncDecl) stmtNode()  {}
 func (n ClassDecl) stmtNode() {}
@@ -1577,17 +1413,22 @@ func (n LiteralExpr) JS() string {
 }
 
 // JSON converts the node back to valid JSON
-func (n LiteralExpr) JSON() (string, error) {
+func (n LiteralExpr) JSON(buf *bytes.Buffer) error {
 	if n.TokenType == TrueToken || n.TokenType == FalseToken || n.TokenType == NullToken || n.TokenType == DecimalToken {
-		return string(n.Data), nil
+		buf.Write(n.Data)
+		return nil
 	} else if n.TokenType == StringToken {
+		data := n.Data
 		if n.Data[0] == '\'' {
-			n.Data[0] = '"'
-			n.Data[len(n.Data)-1] = '"'
+			data = parse.Copy(data)
+			data = bytes.ReplaceAll(data, []byte(`"`), []byte(`\"`))
+			data[0] = '"'
+			data[len(data)-1] = '"'
 		}
-		return string(n.Data), nil
+		buf.Write(data)
+		return nil
 	}
-	return "", ErrInvalidJSON
+	return ErrInvalidJSON
 }
 
 // Element is an array literal element.
@@ -1619,11 +1460,6 @@ func (n Element) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n Element) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ArrayExpr is an array literal.
 type ArrayExpr struct {
 	List []Element
@@ -1669,27 +1505,24 @@ func (n ArrayExpr) JS() string {
 }
 
 // JSON converts the node back to valid JSON
-func (n ArrayExpr) JSON() (string, error) {
-	s := "["
+func (n ArrayExpr) JSON(buf *bytes.Buffer) error {
+	buf.WriteByte('[')
 	for i, item := range n.List {
 		if i != 0 {
-			s += ", "
+			buf.WriteString(", ")
 		}
-		if item.Value != nil {
-			if item.Spread {
-				return "", ErrInvalidJSON
-			}
-			ss, err := item.Value.JSON()
-			if err != nil {
-				return "", err
-			}
-			s += ss
+		if item.Value == nil || item.Spread {
+			return ErrInvalidJSON
+		}
+		val, ok := item.Value.(JSONer)
+		if !ok {
+			return ErrInvalidJSON
+		} else if err := val.JSON(buf); err != nil {
+			return err
 		}
 	}
-	if 0 < len(n.List) && n.List[len(n.List)-1].Value == nil {
-		return "", ErrInvalidJSON
-	}
-	return s + "]", nil
+	buf.WriteByte(']')
+	return nil
 }
 
 // Property is a property definition in an object literal.
@@ -1736,26 +1569,25 @@ func (n Property) JS() string {
 }
 
 // JSON converts the node back to valid JSON
-func (n Property) JSON() (string, error) {
-	s := ""
+func (n Property) JSON(buf *bytes.Buffer) error {
 	if n.Name == nil || n.Name.Literal.TokenType != StringToken && n.Name.Literal.TokenType != IdentifierToken || n.Spread || n.Init != nil {
-		return "", ErrInvalidJSON
+		return ErrInvalidJSON
 	} else if n.Name.Literal.TokenType == IdentifierToken {
-		n.Name.Literal.TokenType = StringToken
-		n.Name.Literal.Data = append(append([]byte{'"'}, n.Name.Literal.Data...), '"')
+		buf.WriteByte('"')
+		buf.Write(n.Name.Literal.Data)
+		buf.WriteByte('"')
+	} else {
+		_ = n.Name.Literal.JSON(buf)
 	}
+	buf.WriteString(": ")
 
-	ss, _ := n.Name.Literal.JSON()
-	s += ss + ": "
-
-	var err error
-	ss, err = n.Value.JSON()
-	if err != nil {
-		fmt.Println("b")
-		return "", err
+	val, ok := n.Value.(JSONer)
+	if !ok {
+		return ErrInvalidJSON
+	} else if err := val.JSON(buf); err != nil {
+		return err
 	}
-	s += ss
-	return s, nil
+	return nil
 }
 
 // ObjectExpr is an object literal.
@@ -1787,19 +1619,18 @@ func (n ObjectExpr) JS() string {
 }
 
 // JSON converts the node back to valid JSON
-func (n ObjectExpr) JSON() (string, error) {
-	s := "{"
+func (n ObjectExpr) JSON(buf *bytes.Buffer) error {
+	buf.WriteByte('{')
 	for i, item := range n.List {
 		if i != 0 {
-			s += ", "
+			buf.WriteString(", ")
 		}
-		ss, err := item.JSON()
-		if err != nil {
-			return "", err
+		if err := item.JSON(buf); err != nil {
+			return err
 		}
-		s += ss
 	}
-	return s + "}", nil
+	buf.WriteByte('}')
+	return nil
 }
 
 // TemplatePart is a template head or middle.
@@ -1817,11 +1648,6 @@ func (n TemplatePart) JS() string {
 	return string(n.Value) + n.Expr.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n TemplatePart) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // TemplateExpr is a template literal or member/call expression, super property, or optional chain with template literal.
 type TemplateExpr struct {
 	Tag  IExpr // can be nil
@@ -1853,11 +1679,6 @@ func (n TemplateExpr) JS() string {
 	return s + string(n.Tail)
 }
 
-// JSON converts the node back to valid JSON
-func (n TemplateExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // GroupExpr is a parenthesized expression.
 type GroupExpr struct {
 	X IExpr
@@ -1872,11 +1693,6 @@ func (n GroupExpr) JS() string {
 	return "(" + n.X.JS() + ")"
 }
 
-// JSON converts the node back to valid JSON
-func (n GroupExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // IndexExpr is a member/call expression, super property, or optional chain with an index expression.
 type IndexExpr struct {
 	X    IExpr
@@ -1893,11 +1709,6 @@ func (n IndexExpr) JS() string {
 	return n.X.JS() + "[" + n.Y.JS() + "]"
 }
 
-// JSON converts the node back to valid JSON
-func (n IndexExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // DotExpr is a member/call expression, super property, or optional chain with a dot expression.
 type DotExpr struct {
 	X    IExpr
@@ -1914,11 +1725,6 @@ func (n DotExpr) JS() string {
 	return n.X.JS() + "." + n.Y.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n DotExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // NewTargetExpr is a new target meta property.
 type NewTargetExpr struct {
 }
@@ -1932,11 +1738,6 @@ func (n NewTargetExpr) JS() string {
 	return "new.target"
 }
 
-// JSON converts the node back to valid JSON
-func (n NewTargetExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ImportMetaExpr is a import meta meta property.
 type ImportMetaExpr struct {
 }
@@ -1950,11 +1751,6 @@ func (n ImportMetaExpr) JS() string {
 	return "import.meta"
 }
 
-// JSON converts the node back to valid JSON
-func (n ImportMetaExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 type Arg struct {
 	Value IExpr
 	Rest  bool
@@ -1977,11 +1773,6 @@ func (n Arg) JS() string {
 	return s + n.Value.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n Arg) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // Args is a list of arguments as used by new and call expressions.
 type Args struct {
 	List []Arg
@@ -2010,11 +1801,6 @@ func (n Args) JS() string {
 	return s
 }
 
-// JSON converts the node back to valid JSON
-func (n Args) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // NewExpr is a new expression or new member expression.
 type NewExpr struct {
 	X    IExpr
@@ -2038,11 +1824,6 @@ func (n NewExpr) JS() string {
 	return "new " + n.X.JS() + "()"
 }
 
-// JSON converts the node back to valid JSON
-func (n NewExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // CallExpr is a call expression.
 type CallExpr struct {
 	X    IExpr
@@ -2058,11 +1839,6 @@ func (n CallExpr) JS() string {
 	return n.X.JS() + "(" + n.Args.JS() + ")"
 }
 
-// JSON converts the node back to valid JSON
-func (n CallExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // OptChainExpr is an optional chain.
 type OptChainExpr struct {
 	X IExpr
@@ -2094,11 +1870,6 @@ func (n OptChainExpr) JS() string {
 	}
 }
 
-// JSON converts the node back to valid JSON
-func (n OptChainExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // UnaryExpr is an update or unary expression.
 type UnaryExpr struct {
 	Op TokenType
@@ -2125,8 +1896,13 @@ func (n UnaryExpr) JS() string {
 }
 
 // JSON converts the node back to valid JSON
-func (n UnaryExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
+func (n UnaryExpr) JSON(buf *bytes.Buffer) error {
+	if lit, ok := n.X.(*LiteralExpr); ok && n.Op == NegToken && lit.TokenType == DecimalToken {
+		buf.WriteByte('-')
+		buf.Write(lit.Data)
+		return nil
+	}
+	return ErrInvalidJSON
 }
 
 // BinaryExpr is a binary expression.
@@ -2147,11 +1923,6 @@ func (n BinaryExpr) JS() string {
 	return n.X.JS() + " " + n.Op.String() + " " + n.Y.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n BinaryExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // CondExpr is a conditional expression.
 type CondExpr struct {
 	Cond, X, Y IExpr
@@ -2166,11 +1937,6 @@ func (n CondExpr) JS() string {
 	return n.Cond.JS() + " ? " + n.X.JS() + " : " + n.Y.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n CondExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // YieldExpr is a yield expression.
 type YieldExpr struct {
 	Generator bool
@@ -2200,11 +1966,6 @@ func (n YieldExpr) JS() string {
 	return s + " " + n.X.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n YieldExpr) JSON() (string, error) {
-	return "", ErrInvalidJSON
-}
-
 // ArrowFunc is an (async) arrow function.
 type ArrowFunc struct {
 	Async  bool
@@ -2229,9 +1990,32 @@ func (n ArrowFunc) JS() string {
 	return s + n.Params.JS() + " => " + n.Body.JS()
 }
 
-// JSON converts the node back to valid JSON
-func (n ArrowFunc) JSON() (string, error) {
-	return "", ErrInvalidJSON
+// CommaExpr is a series of comma expressions.
+type CommaExpr struct {
+	List []IExpr
+}
+
+func (n CommaExpr) String() string {
+	s := "("
+	for i, item := range n.List {
+		if i != 0 {
+			s += ","
+		}
+		s += item.String()
+	}
+	return s + ")"
+}
+
+// JS converts the node back to valid JavaScript
+func (n CommaExpr) JS() string {
+	s := ""
+	for i, item := range n.List {
+		if i != 0 {
+			s += ","
+		}
+		s += item.JS()
+	}
+	return s
 }
 
 func (v *Var) exprNode()           {}
@@ -2252,3 +2036,4 @@ func (n BinaryExpr) exprNode()     {}
 func (n CondExpr) exprNode()       {}
 func (n YieldExpr) exprNode()      {}
 func (n ArrowFunc) exprNode()      {}
+func (n CommaExpr) exprNode()      {}
